@@ -12,7 +12,7 @@ use Math::Prime::Util qw/ prime_get_config
 
 BEGIN {
   $Math::Prime::Util::RandomPrimes::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::RandomPrimes::VERSION = '0.37';
+  $Math::Prime::Util::RandomPrimes::VERSION = '0.52';
 }
 
 BEGIN {
@@ -242,19 +242,21 @@ sub get_randf_nbit {
 # possibility is to, if they do not supply a rand function, use the GMP MT
 # function with an appropriate seed.
 #
-# Random timings for 10M calls:
-#    1.92    system rand
-#    2.62    Math::Random::MT::Auto
-#   12.0     Math::Random::Secure           w/ISAAC::XS
-#   12.6     Bytes::Random::Secure OO       w/ISAAC::XS     <==== our
-#   31.1     Bytes::Random::Secure OO                       <==== default
-#   44.5     Bytes::Random::Secure function w/ISAAC::XS
-#   44.8     Math::Random::Secure
-#   71.5     Bytes::Random::Secure function
-# 1840       Crypt::Random
+# Random timings for 10M calls on i4770K:
+#    0.39   Math::Random::MTwist 0.13
+#    0.89   system rand
+#    1.76   Math::Random::MT::Auto
+#    5.35   Bytes::Random::Secure OO       w/ISAAC::XS     <==== our
+#    7.43   Math::Random::Secure           w/ISAAC::XS
+#   12.40   Math::Random::Secure
+#   12.78   Bytes::Random::Secure OO                       <==== default
+#   13.86   Bytes::Random::Secure function w/ISAAC::XS
+#   21.95   Bytes::Random::Secure function
+#  822.1    Crypt::Random
 #
+# time perl -E 'use Math::Random::MTwist "irand32"; irand32() for 1..10000000;'
 # time perl -E 'sub irand {int(rand(4294967296));} irand() for 1..10000000;'
-# time perl -E 'use Math::Random::MT::Auto qw/irand/; irand() for 1..10000000;'
+# time perl -E 'use Math::Random::MT::Auto; sub irand { Math::Random::MT::Auto::irand() & 0xFFFFFFFF } irand() for 1..10000000;'
 # time perl -E 'use Math::Random::Secure qw/irand/; irand() for 1..10000000;'
 # time perl -E 'use Bytes::Random::Secure qw/random_bytes/; sub irand {return unpack("L",random_bytes(4));} irand() for 1..10000000;'
 # time perl -E 'use Bytes::Random::Secure; my $rng = Bytes::Random::Secure->new(); sub irand {return $rng->irand;} irand() for 1..10000000;'
@@ -294,6 +296,17 @@ my $_random_prime = sub {
     $low = Math::BigInt->new("$low") if ref($high) eq 'Math::BigInt';
     confess "Invalid _random_prime parameters: $low, $high" if ($low % 2) == 0 || ($high % 2) == 0;
 
+    # If they want to use the crappy PRIMEINC algorithm, do it.
+    if (prime_get_config->{'use_primeinc'}) {
+      $prime = $low + $_RANDF->($high-$low);   # Random number from low-high
+      $prime = next_prime($prime-1);           # Increment until prime
+      if ($prime > $high) {                    # Rollover
+        $prime = next_prime($low-1);
+        return if $prime > $high;
+      }
+      return $prime;
+    }
+
     # We're going to look at the odd numbers only.
     my $oddrange = (($high - $low) >> 1) + 1;
 
@@ -501,8 +514,7 @@ sub random_prime {
 
   # Tighten the range to the nearest prime.
   $low = ($low <= 2)  ?  2  :  next_prime($low-1);
-  # TODO: if high is bigint, we should do high++?
-  $high = ($high < ~0)  ?  prev_prime($high + 1)  :  prev_prime($high);
+  $high = ($high == ~0) ? prev_prime($high) : prev_prime($high + 1);
   return $low if ($low == $high) && is_prob_prime($low);
   return if $low >= $high;
 
@@ -560,7 +572,7 @@ sub random_nbit_prime {
   _set_randf();
 
   # Very small size, use the nth-prime method
-  if ($bits <= 18 && MPU_USE_XS) {
+  if ($bits <= 20 && MPU_USE_XS) {
     if ($bits <= 4) {
       return (2,3)[$_RANDF_NBIT->(1)] if $bits == 2;
       return (5,7)[$_RANDF_NBIT->(1)] if $bits == 3;
@@ -572,6 +584,22 @@ sub random_nbit_prime {
   croak "Mid-size random primes not supported on broken old Perl"
     if OLD_PERL_VERSION && MPU_64BIT && $bits > 49 && $bits <= 64;
 
+  # If they want to use the crappy PRIMEINC algorithm, do it.
+  if (prime_get_config->{'use_primeinc'}) {
+    my($prime, $low, $high);
+    $low  = ($bits >  MPU_MAXBITS) ? Math::BigInt->bone->blsft($bits-1)
+                                   : (1 << ($bits-1));
+    $high = ($bits >  MPU_MAXBITS) ? Math::BigInt->bone->blsft($bits)->bdec()
+          : ($bits == MPU_MAXBITS) ? ~0
+                                   : (1 << $bits) - 1;
+    $prime = next_prime($low + $_RANDF_NBIT->($bits-1) - 1);
+    if ($prime > $high) {                    # Rollover
+      $prime = next_prime($low-1);
+      return if $prime > $high;
+    }
+    return $prime;
+  }
+
   # Fouque and Tibouchi (2011) Algorithm 1 (basic)
   # Modified to make sure the nth bit is always set.
   #
@@ -832,6 +860,146 @@ sub random_maurer_prime_with_cert {
   croak "Failure in random_maurer_prime, could not find a prime\n";
 } # End of random_maurer_prime
 
+
+sub random_shawe_taylor_prime_with_cert {
+  my $k = shift;
+
+  my $seed;
+  my $irandf = prime_get_config->{'irand'};
+  if (!defined $irandf) {
+    if (!defined $_BRS) {
+      require Bytes::Random::Secure;
+      $_BRS = Bytes::Random::Secure->new(NonBlocking=>1);
+    }
+    $seed = $_BRS->bytes(512/8);
+  } else {
+    $seed = pack("L*", map { $irandf->() } 0 .. (512>>5));
+  }
+
+  my($status,$prime,$prime_seed,$prime_gen_counter,$cert)
+     = _ST_Random_prime($k, $seed);
+  croak "Shawe-Taylor random prime failure" unless $status;
+  croak "Shawe-Taylor random prime failure: prime $prime failed certificate verification!" unless verify_prime($cert);
+
+  return ($prime,$cert);
+}
+
+sub _seed_plus_one {
+    my($s) = @_;
+    for (my $i = length($s)-1; $i >= 0; $i--) {
+        vec($s, $i, 8)++;
+        last unless vec($s, $i, 8) == 0;
+    }
+    return $s;
+}
+
+sub _ST_Random_prime {  # From FIPS 186-4
+  my($k, $input_seed) = @_;
+  croak "Shawe-Taylor random prime must have length >= 2"  if $k < 2;
+  $k = int("$k");
+
+  croak "Shawe-Taylor random prime, invalid input seed"
+     unless defined $input_seed && length($input_seed) >= 32;
+
+  if (!defined $Digest::SHA::VERSION) {
+    eval { require Digest::SHA;
+           my $version = $Digest::SHA::VERSION;
+           $version =~ s/[^\d.]//g;
+           $version >= 4.00; }
+      or do { croak "Must have Digest::SHA 4.00 or later"; };
+  }
+
+  my $k2 = Math::BigInt->new(2)->bpow($k-1);
+
+  if ($k < 33) {
+    my $seed = $input_seed;
+    my $prime_gen_counter = 0;
+    my $kmask    = 0xFFFFFFFF >> (32-$k);    # Does the mod operation
+    my $kstencil = (1 << ($k-1)) | 1;        # Sets high and low bits
+    while (1) {
+      my $seedp1 = _seed_plus_one($seed);
+      my $cvec = Digest::SHA::sha256($seed) ^ Digest::SHA::sha256($seedp1);
+      # my $c = Math::BigInt->from_hex('0x' . unpack("H*", $cvec));
+      # $c = $k2 + ($c % $k2);
+      # $c = (2 * ($c >> 1)) + 1;
+      my($c) = unpack("N*", substr($cvec,-4,4));
+      $c = ($c & $kmask) | $kstencil;
+      $prime_gen_counter++;
+      $seed = _seed_plus_one($seedp1);
+      my ($isp, $cert) = is_provable_prime_with_cert($c);
+      return (1,$c,$seed,$prime_gen_counter,$cert) if $isp;
+      return (0,0,0,0) if $prime_gen_counter > 10000 + 16*$k;
+    }
+  }
+  my($status,$c0,$seed,$prime_gen_counter,$cert)
+     = _ST_Random_prime( (($k+1)>>1)+1, $input_seed);
+  return (0,0,0,0) unless $status;
+  $cert = ($c0 < Math::BigInt->new("18446744073709551615"))
+          ? "" : _strip_proof_header($cert);
+  my $iterations = int(($k + 255) / 256) - 1;  # SHA256 generates 256 bits
+  my $old_counter = $prime_gen_counter;
+  my $xstr = '';
+  for my $i (0 .. $iterations) {
+    $xstr = Digest::SHA::sha256_hex($seed) . $xstr;
+    $seed = _seed_plus_one($seed);
+  }
+  my $x = Math::BigInt->from_hex('0x'.$xstr);
+  $x = $k2 + ($x % $k2);
+  my $t = ($x + 2*$c0 - 1) / (2*$c0);
+  _make_big_gcds() if $_big_gcd_use < 0;
+  while (1) {
+    if (2*$t*$c0 + 1 > 2*$k2) { $t = ($k2 + 2*$c0 - 1) / (2*$c0); }
+    my $c = 2*$t*$c0 + 1;
+    $prime_gen_counter++;
+
+    # Don't do the Pocklington check unless the candidate looks prime
+    my $looks_prime = 0;
+    if (MPU_USE_GMP) {
+      # MPU::GMP::is_prob_prime has fast tests built in.
+      $looks_prime = Math::Prime::Util::GMP::is_prob_prime($c);
+    } else {
+      # No GMP, so first do trial divisions, then a SPSP test.
+      $looks_prime = Math::BigInt::bgcd($c, 111546435)->is_one;
+      if ($looks_prime && $_big_gcd_use && $c > $_big_gcd_top) {
+        $looks_prime = Math::BigInt::bgcd($c, $_big_gcd[0])->is_one &&
+                       Math::BigInt::bgcd($c, $_big_gcd[1])->is_one &&
+                       Math::BigInt::bgcd($c, $_big_gcd[2])->is_one &&
+                       Math::BigInt::bgcd($c, $_big_gcd[3])->is_one;
+      }
+      $looks_prime = 0 if $looks_prime && !is_strong_pseudoprime($c, 3);
+    }
+
+    if ($looks_prime) {
+      # We could use a in (2,3,5,7,11,13), but pedantically use FIPS 186-4.
+      my $astr = '';
+      for my $i (0 .. $iterations) {
+        $astr = Digest::SHA::sha256_hex($seed) . $astr;
+        $seed = _seed_plus_one($seed);
+      }
+      my $a = Math::BigInt->from_hex('0x'.$astr);
+      $a = ($a % ($c-3)) + 2;
+      my $z = $a->copy->bmodpow(2*$t,$c);
+      if (Math::BigInt::bgcd($z-1,$c)->is_one && $z->copy->bmodpow($c0,$c)->is_one) {
+        croak "Shawe-Taylor random prime failure at ($k): $c not prime"
+          unless is_prob_prime($c);
+        $cert = "[MPU - Primality Certificate]\nVersion 1.0\n\n" .
+                 "Proof for:\nN $c\n\n" .
+                 "Type Pocklington\nN $c\nQ $c0\nA $a\n" .
+                 $cert;
+        return (1, $c, $seed, $prime_gen_counter, $cert);
+      }
+    } else {
+      # Update seed "as if" we performed the Pocklington check from FIPS 186-4
+      for my $i (0 .. $iterations) {
+        $seed = _seed_plus_one($seed);
+      }
+    }
+    return (0,0,0,0) if $prime_gen_counter > 10000 + 16*$k + $old_counter;
+    $t++;
+  }
+}
+
+
 # Gordon's algorithm for generating a strong prime.
 sub random_strong_prime {
   my $t = shift;
@@ -937,7 +1105,7 @@ Math::Prime::Util::RandomPrimes - Generate random primes
 
 =head1 VERSION
 
-Version 0.37
+Version 0.52
 
 
 =head1 SYNOPSIS
@@ -986,6 +1154,18 @@ Construct a random provable prime of C<n> bits using Maurer's FastPrime
 algorithm.  C<n> must be at least 2.  Returns a list of two items: the
 prime and the certificate.
 
+=head2 random_shawe_taylor_prime
+
+Construct a random provable prime of C<n> bits using Shawe-Taylor's
+algorithm.  C<n> must be at least 2.  The implementation is from
+FIPS 186-4 and uses SHA-256 with 512 bits of randomness.
+
+=head2 random_shawe_taylor_prime_with_cert
+
+Construct a random provable prime of C<n> bits using Shawe-Taylor's
+algorithm.  C<n> must be at least 2.  Returns a list of two items: the
+prime and the certificate.
+
 =head2 random_strong_prime
 
 Construct a random strong prime of C<n> bits.  C<n> must be at least 128.
